/*
 * Author:
 *    Mark Phillips
 *    University of North Carolina Asheville /  National Climatic Data Center
 *    Mark.Phillips@noaa.gov
 *    828-271-4487
 */

package gov.noaa.ncdc.swdi.gridder;

/**

* The idea behind DomainGridder is simple: provide a convenient way
* to convert back and forth between a range of floating point numbers
* and a finite number of equally spaced sample points in that range.
* This seems almost trivial at first --- just round off to the
* nearest sample point.  In practice, however, it's more compilicated
* than that, which is why it merits an entire Java class to handle
* it.
*
* <p>
*
* To be more precise, think of the problem as follows: suppose you
* have a "grid" of N equally spaced floating point numbers ranging
* from MIN to MAX.  Number these grid points 0 through N-1.  You want
* to provide a way to convert a floating point number to the grid
* point "closest" to it, and a way to convert the grid numbers 0
* through N-1 to their floating point values.
*
* <p>
*
* What's really going on here is that a range of floating point
* numbers is being divided into equal-length subintervals, or cells.
* There is one cell per grid point, and a floating point number is
* mapped to the grid point corresponding to the cell it lies in.
*
* <p>
*
* In theory there are an infinite number of ways that the grid cells can
* be aligned with respect to the grid points.  In practice, however,
* there are only three that are commonly used: the grid point may be
* at the left endpoint of its cell, in the center, or at the right endpoint.
*
* <p>
*
* When a DomainGridder instance is created, the following things are
* specified: the minimum and maximum grid points, the grid cell
* length, and an alignment (one of the keywords MIN, MAX, or CENTER).
* The relationship between the grid points and the grid cells for
* each of the three alignment possibilities is depicted in the
* diagram below.  In the diagram, the grid points are marked with
* 'P', and the grid cell boundaries with the vertical bar symbol
* ('|').
*
* <pre>
*
* alignment=MIN:
*                 domainMin                           domainMax
*                 P           P           P           P
*                 |           |           |           |           |
*                 | cell 0    |  cell 1   |  ...      | last cell |
*
*
* alignment=MAX:
*                 domainMin                           domainMax
*                 P           P           P           P
*     |           |           |           |           |
*     | cell 0    |  cell 1   |  ...      | last cell |
*
*
* alignment=CENTER:
*                 domainMin                           domainMax
*                 P           P           P           P
*           |           |           |           |           |
*           | cell 0    |  cell 1   |  ...      | last cell |
*
* </pre>
*
* DomainGridder assumes that the length of the domain is an integer
* multiple of the cell length --- in other words, that the domain
* will divide into an exact integer number of grid cells.  Its
* behavior is not defined if this is not the case.
*
* A DomainGridder instance provides methods for retrieving the
* original input values, the number of cells in the grid, and for
* converting betweend domain values and grid index values.
*
* Optionally, a floating point precision value may be specified when
* a DomainGridder is created.  This precision is used only when
* converting floating point domain values near the domain endpoints
* to cell indices; domain values that are within precision of the
* left endpoint of the 0th cell are considered to be in that cell,
* and domain values that are within precision of the right endpoint
* of the last cell are considered to be in that cell.  If precision
* is not specified, a precision of 0.0 is assumed.
*
* @author Mark Phillips Wed Jul 11 14:58:21 2007
*/
public class DomainGridder {

    /**
     * The Alignment enum is used to specify how grid cells are aligned
     * with respect to the corresponding grid points
     * <ol>
     *   <li>MIN means that each grid point is the left endpoint
     *       of the corresponding cell
     *   <li>MAX means that each grid point is the right endpoint
     *       of the corresponding cell
     *   <li>CENTER means that each grid point is the center
     *       point of the corresponding cell
     *   <li>BOUNDS is experimental and will probably
     *       be removed soon.  Don't use it.
     * </ol>
     */
    public static enum Alignment { MIN, CENTER, MAX, BOUNDS }

    public static class UnknownAlignmentException extends Exception {}

    public static Alignment parseAlignment(String s) throws UnknownAlignmentException {
		s = s.toUpperCase();
		if (s.equals("MIN"))    { return DomainGridder.Alignment.MIN; }
		if (s.equals("MAX"))    { return DomainGridder.Alignment.MAX; }
		if (s.equals("CENTER")) { return DomainGridder.Alignment.CENTER; }
		if (s.equals("BOUNDS")) { return DomainGridder.Alignment.BOUNDS; }
		throw new UnknownAlignmentException();
    }

    public static class OutOfBoundsException extends Exception {
		public OutOfBoundsException(String msg) {
			super(msg);
		}
		public OutOfBoundsException() {
			super();
		}
    }

    private Alignment cellAlignment;
    private double precision;
    private double domainMin, domainMax, domainLength, cellLength;
    private int cellCount, indexMin, indexMax;

    private double domainMinAccept, domainMaxAccept;

    /**
     * Create a new DomainGridder instance with a given cell length.
     *
     * @param domainMin	        The minimum endpoint of the domain.
     * @param domainMax	        The maximum endpoint of the domain.
     * @param cellLength	The length of a grid cell.
     * @param cellAlignment	The cell alignment.
     * @param precision         Precision to be used at the domain endpoints.
     */
    public DomainGridder(double domainMin, double domainMax, double cellLength, Alignment cellAlignment, double precision) {
		this.domainMin     = domainMin;
		this.domainMax     = domainMax;
		this.cellLength    = cellLength;
		this.cellAlignment = cellAlignment;
		this.precision     = precision;

		this.domainLength = domainMax - domainMin;
		this.cellCount    = (int)Math.round( domainLength / cellLength ) + ((cellAlignment != Alignment.BOUNDS) ? 1 : 0);
		this.indexMin     = 0;
		this.indexMax     = cellCount - 1;

		switch (cellAlignment) {
		case MIN:
			this.domainMinAccept = domainMin;
			this.domainMaxAccept = domainMax + cellLength;
			break;
		case CENTER:
			this.domainMinAccept = domainMin - cellLength/2;
			this.domainMaxAccept = domainMax + cellLength/2;
			break;
		case MAX:
			this.domainMinAccept = domainMin - cellLength;
			this.domainMaxAccept = domainMax;
			break;
		case BOUNDS:
			this.domainMinAccept = domainMin;
			this.domainMaxAccept = domainMax;
			break;
		}
    }

    /**
     * Create a new DomainGridder instance with a given cell length, and a precision of 0.0.
     *
     * @param domainMin	        The minimum endpoint of the domain.
     * @param domainMax	        The maximum endpoint of the domain.
     * @param cellLength	The length of a grid cell.
     * @param cellAlignment	The cell alignment.
     */
    public DomainGridder(double domainMin, double domainMax, double cellLength, Alignment cellAlignment) {
		this(domainMin, domainMax, cellLength, cellAlignment, 0.0);
    }

    /**
     * Create a new DomainGridder instance with a given number of cells; the cell length
     * is computed as (domainMax - domainMin) / cellCount.
     *
     * @param domainMin	        The minimum endpoint of the domain.
     * @param domainMax	        The maximum endpoint of the domain.
     * @param cellCount	        The number of cells in the grid.
     * @param cellAlignment	The cell alignment.
     * @param precision         Precision to be used at the domain endpoints.
     */
    public DomainGridder(double domainMin, double domainMax, int cellCount, Alignment cellAlignment, double precision) {
		this(domainMin, domainMax, (domainMax - domainMin) / cellCount, cellAlignment, precision);
    }

    /**
     * Create a new DomainGridder instance with a given number of cells, and a precision of 0.0.  The cell length
     * is computed as (domainMax - domainMin) / cellCount.
     *
     * @param domainMin	        The minimum endpoint of the domain.
     * @param domainMax	        The maximum endpoint of the domain.
     * @param cellCount	        The number of cells in the grid.
     * @param cellAlignment	The cell alignment.
     */
    public DomainGridder(double domainMin, double domainMax, int cellCount, Alignment cellAlignment) {
		this(domainMin, domainMax, cellCount, cellAlignment, 0.0);
    }

    /**
     * Convert a value in the domain to an integer cell index. Throws an OutOfBoundsException
     * if the input value is less than domainMin - precision or greater than domainMax + precision.
     */
    public int domainToIndex(double d) throws OutOfBoundsException {
		if (d < domainMinAccept) {
			if (d < domainMinAccept - precision) {
				throw new OutOfBoundsException("value out of bounds");
			} else {
				d += precision;
			}
		}
		if (cellAlignment == Alignment.BOUNDS) {
			if (d > domainMaxAccept) {
				if (d > domainMaxAccept + precision) {
					throw new OutOfBoundsException("value out of bounds");
				} else {
					d -= precision;
				}
			}
		} else {
			if (d >= domainMaxAccept) {
				if (d >= domainMaxAccept + precision) {
					throw new OutOfBoundsException("value out of bounds");
				} else {
					d -= precision;
				}
			}
		}
		return (int)Math.round(Math.floor( (d - domainMinAccept) / cellLength ));
    }

    /**
     * Convert an integer cell index to its corresponding domain value.  Throws an OutOfBoundsException
     * if the input index is less than 0 or greater than or equal to cellCount.
     */
    public double indexToDomain(int i) throws OutOfBoundsException {
		if (i < 0 || i > indexMax) {
			throw new OutOfBoundsException("index out of bounds");
		}
		return i * domainLength / indexMax + domainMin;
    }

    /**
     * Return the domain minimum (left) endpoint.
     */
    public double getDomainMin()    { return this.domainMin;    }

    /**
     * Return the domain maximum (right) endpoint.
     */
    public double getDomainMax()    { return this.domainMax;    }

    /**
     * Return the length of the domain; this is the same
     * as domainMax - domainMin.
     */
    public double getDomainLength() { return this.domainLength; }

    /**
     * Return the cell length.
     */
    public double getCellLength()   { return this.cellLength;   }

    /**
     * Return the number of cells in the domain.
     */
    public int    getCellCount()    { return this.cellCount;    }

    /**
     * Return the minimum cell index value.  This is currently always 0, but
     * future versions of DomainGridder() might be modified to work with
     * non-zero-based indexing.
     */
    public int    getIndexMin()     { return this.indexMin;     }

    /**
     * Return the maximum cell index value; this is the same as getCellCount() - 1.
     */
    public int    getIndexMax()     { return this.indexMax;     }


    public static void main(String argv[]) throws Exception {

		int argi = 0;
		double d_min  = Double.parseDouble(argv[argi++]);
		double d_max  = Double.parseDouble(argv[argi++]);
		double d_cell = Double.parseDouble(argv[argi++]);
		DomainGridder.Alignment alignment = DomainGridder.parseAlignment(argv[argi++]);

		System.out.printf(" [ %3.1f, %3.1f ] / %3.1f { %s }:\n", d_min, d_max, d_cell, alignment);
		DomainGridder dg = new DomainGridder(d_min, d_max, d_cell, alignment, 0.01);
		System.out.printf("cell count = %1d\n", dg.getCellCount());
		for (double d = d_min - 2*d_cell; d <= d_max + 2*d_cell; d += 0.1) {
			int i = -9999;
			try {
				i = dg.domainToIndex(d);
			} catch (DomainGridder.OutOfBoundsException e) {
				System.out.printf(" %3.1f ==> out of bounds\n", d);
				continue;
			}
			System.out.printf(" %4.1f ==> [%2d]\n", d, i);
		}
		for (int i=dg.getIndexMin()-1; i<=dg.getIndexMax()+1; ++i) {
			double d = -9999;
			try {
				d = dg.indexToDomain(i);
			} catch (DomainGridder.OutOfBoundsException e) {
				System.out.printf(" [%2d] <== out of bounds\n", i);
				continue;
			}
			System.out.printf(" [%2d] <== %4.1f\n", i, d);
		}

    }

}
