package gov.noaa.ncdc.swdi.gridder;

import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.data.FeatureSource;
import org.geotools.data.shapefile.ShapefileDataStore;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.SchemaException;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.map.DefaultMapLayer;
import org.geotools.map.MapContext;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.geotools.styling.RasterSymbolizer;
import org.geotools.styling.SLDParser;
import org.geotools.styling.Style;
import org.geotools.styling.StyleBuilder;
import org.geotools.styling.StyleFactory;
import org.geotools.styling.StyleFactoryFinder;
import org.opengis.feature.IllegalAttributeException;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.TransformException;

import com.vividsolutions.jts.geom.Envelope;

/**
 * This class provides streamlined and easy abstraction for several WRT common actions.  
 * @author steve.ansari
 *
 */
public class GTHelper {



    public static ReferencedEnvelope getReferencedEnvelope(GridCoverage2D gc, MapContext context) {
        org.opengis.geometry.Envelope env = gc.getEnvelope();

        //        System.out.println("GC ENV: " + env);

        ReferencedEnvelope refEnv = new ReferencedEnvelope(new Envelope(
                env.getLowerCorner().getOrdinate(0), // X1
                env.getUpperCorner().getOrdinate(0), // X2
                env.getLowerCorner().getOrdinate(1), // Y1
                env.getUpperCorner().getOrdinate(1)), // Y2
                context.getCoordinateReferenceSystem());

        //        System.out.println("GC REFERENCED ENV: " + refEnv);

        return refEnv;

    }


    /**
     * Adds a shapefile to a map context given the URL and URL of SLD style file
     * 
     * @param context
     * @param shape
     * @param sldURL
     * @param title
     * @return The bounds of this shapefile
     * @throws MalformedURLException
     * @throws IOException
     */
    public static ReferencedEnvelope addShapefileToContext(MapContext context, URL shape, URL sldURL, String title) 
    throws MalformedURLException, IOException {

        //IndexedShapefileDataStore ds = new IndexedShapefileDataStore(shape);
        ShapefileDataStore ds = new ShapefileDataStore(shape);

        //        MemoryDataStore mds = new MemoryDataStore();
//        System.out.println("WRTHELPER: LOADING SHP: " + shape);
//        System.out.println("WRTHELPER: LOADING SLD: " + sldURL);

        FeatureSource fs = ds.getFeatureSource();
        //        mds.addFeatures(fs.getFeatures());

        Envelope env = fs.getBounds();
//        System.out.println(fs.getFeatures().size() + " FEATURES -- " + env);
        StyleFactory factory = StyleFactoryFinder.createStyleFactory();
        SLDParser stylereader = new SLDParser(factory, sldURL);
        org.geotools.styling.Style[] style = stylereader.readXML();

        context.addLayer(new DefaultMapLayer(fs.getFeatures(), style[0], title));
        //  context.addLayer(new DefaultMapLayer(mds.get.getFeatureSource().getFeatures(), style[0], title));

        //return new ReferencedEnvelope(env, context.getCoordinateReferenceSystem());
        return new ReferencedEnvelope(env, DefaultGeographicCRS.WGS84);

    }





    /**
     * Adjusts the geographic bounds equally in North,South,East,West so image 
     * always has equal x and y scale and is not skewed.
     * @param image
     * @param bounds
     * @return
     */
    public static ReferencedEnvelope adjustGeographicBounds(BufferedImage image, ReferencedEnvelope bounds) {
        return adjustGeographicBounds(new Dimension(image.getWidth(), image.getHeight()), bounds); 
    }

    /**
     * Adjusts the geographic bounds equally in North,South,East,West so image 
     * always has equal x and y scale and is not skewed.
     * @param image
     * @param bounds
     * @return
     */
    public static ReferencedEnvelope adjustGeographicBounds(Dimension imageDimension, java.awt.geom.Rectangle2D bounds, CoordinateReferenceSystem crs) {

        ReferencedEnvelope refEnv = getReferencedEnvelope(bounds, crs);
        return adjustGeographicBounds(imageDimension, refEnv);
    }       

    /**
     * Adjusts the geographic bounds equally in North,South,East,West so image 
     * always has equal x and y scale and is not skewed.
     * @param imageDimension
     * @param bounds
     * @return
     */
    public static ReferencedEnvelope adjustGeographicBounds(Dimension imageDimension, ReferencedEnvelope bounds) {

        double imgRatio = (double)imageDimension.getWidth() / (double)imageDimension.getHeight();
        double geoRatio = bounds.getWidth() / bounds.getHeight();

        double dlon = bounds.getWidth();
        double dlat = bounds.getHeight();
        double geoCenterX = bounds.getMinX() + (dlon / 2.0);
        double geoCenterY = bounds.getMinY() + (dlat / 2.0);

        double x = imgRatio / geoRatio;

        if (x > 1.0) {
            dlon = dlon * x;
        }
        else {
            dlat = dlat / x;
        }

        return new ReferencedEnvelope(geoCenterX + dlon / 2.0, 
                geoCenterX - dlon / 2.0, 
                geoCenterY + dlat / 2.0, 
                geoCenterY - dlat / 2.0,
                bounds.getCoordinateReferenceSystem());
    }  //  end adjustGeographicBounds()


    public static ReferencedEnvelope adjustGeographicBounds(Dimension imageDimension, double geoCenterX, double geoCenterY, 
            double geoWidth, double geoHeight, CoordinateReferenceSystem crs) {

        ReferencedEnvelope refEnv = new ReferencedEnvelope(
                geoCenterX-geoWidth/2.0,  // x1 
                geoCenterX+geoWidth/2.0,  // x2
                geoCenterY-geoHeight/2.0, // y1
                geoCenterY+geoHeight/2.0, // y2
                crs);

        return adjustGeographicBounds(imageDimension, refEnv);
    }

    public static ReferencedEnvelope getReferencedEnvelope(java.awt.geom.Rectangle2D bounds, CoordinateReferenceSystem crs) {
        ReferencedEnvelope refEnv = new ReferencedEnvelope(
                bounds.getMinX(), // X1
                bounds.getMaxX(), // X2
                bounds.getMinY(), // Y1
                bounds.getMaxY(), // Y2
                crs);
        return refEnv;
    }










    /**
     * Add generic feature collection to the context using a SLD URL
     * @param context
     * @param features
     * @param sldURL
     * @return
     * @throws IOException
     * @throws FactoryException
     */
    public static ReferencedEnvelope addFeaturesToContext(MapContext context, FeatureCollection features, URL sldURL) 
    throws IOException, FactoryException {

        StyleFactory factory = StyleFactoryFinder.createStyleFactory();
        SLDParser stylereader = new SLDParser(factory, sldURL);
        org.geotools.styling.Style[] style = stylereader.readXML();
        return addFeaturesToContext(context, features, style[0]);

    }

    /**
     * Add generic feature collection to the context using a Style
     * @param context
     * @param features
     * @param style
     * @return
     * @throws IOException
     * @throws FactoryException
     */
    public static ReferencedEnvelope addFeaturesToContext(MapContext context, FeatureCollection features, Style style) 
    throws IOException, FactoryException {

//        System.out.println("ADDING FEATURES: "+features.getSchema().getName());

        context.addLayer(new DefaultMapLayer(features, style, features.getSchema().getName().toString()));
        return new ReferencedEnvelope(features.getBounds(), context.getCoordinateReferenceSystem());
    }




    public static ReferencedEnvelope addGridCoverageToContext(MapContext context, GridCoverage2D gc) 
    throws SchemaException, TransformException, IllegalAttributeException {

        final StyleBuilder sb = new StyleBuilder();
        final RasterSymbolizer rs = sb.createRasterSymbolizer();
        Style style = sb.createStyle(rs);

        context.addLayer(new DefaultMapLayer(gc, style, gc.getName().toString()));

        return getReferencedEnvelope(gc, context); 
    }







}
