/*
 * Author:
 *    Mark Phillips
 *    University of North Carolina Asheville /  National Climatic Data Center
 *    Mark.Phillips@noaa.gov
 *    828-271-4487
 */

/*
 * usage:
 *    -w OUTPUT_IMAGE_WIDTH<int>
 *    -n NUMCLASSES<int>   (use N classes evenly spaced to max data value)
 *    -m MAX_VALUE<double> (use given value as max data value for use with -n)
 *    -N NUMCLASSES<int>   (use N classes which evenly distribute the data; overrides -n and -m options)
 *    -C v0,v1,v2,...,vN   (list of endpoints for N-1 classes; overrides -n and -m and -N options)
 *    -s SCALEFACTOR       (multiply data values by this scale factor)
 *    -t TITLE<string>
 *    -nl (do not draw legend)
 *    -h (include the counts (histogram) under the legend)
 *    -g (add grid cell size annotation to bottom right corner of image)
 *    -a X<int> Y<int> ANNOTATION<string>
 *    -lt<string> (add legend title; multiple allowed)
 *    -lp X<int> Y<int> (legend position)
 *    -lf LEGEND_PRINTF_FORMAT<string>
 *    -c R1<int> G1<int> B1<int> R2<int> G2<int> B2<int>
 */


// http://docs.codehaus.org/display/GEOTDOC/02+JTS+Utility+Class:
// 
//     import org.geotools.geometry.jts.JTS;
//     import org.geotools.referencing.CRS;
// 
//     CoordinateReferenceSystem sourceCRS = CRS.decode("EPSG:4326");
//     CoordinateReferenceSystem targetCRS = CRS.decode("EPSG:23032");
// 
//     MathTransform transform = CRS.findMathTransform(sourceCRS, targetCRS);
//     Geometry targetGeometry = JTS.transform( sourceGeometry, transform);


package gov.noaa.ncdc.swdi.gridder;

import org.geotools.geometry.jts.JTS;
import org.geotools.referencing.CRS;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.Coordinate;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.MathTransform;

import java.awt.AlphaComposite;
import java.awt.Font;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.awt.image.ImageObserver;
import java.awt.Image;
import java.io.File;
import java.util.List;
import java.util.ArrayList;    

import java.awt.Rectangle;
import java.awt.font.FontRenderContext;
import java.awt.font.GlyphVector;
import java.awt.geom.Rectangle2D;
import java.net.URL;

import javax.imageio.ImageIO;

import org.geotools.feature.SchemaException;
import org.geotools.geometry.GeneralEnvelope;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.map.DefaultMapContext;
import org.geotools.map.MapContext;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.geotools.renderer.lite.StreamingRenderer;
import org.opengis.feature.IllegalAttributeException;
import org.opengis.referencing.operation.TransformException;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridCoverageFactory;

import com.vividsolutions.jts.geom.Envelope;

import ucar.nc2.NetcdfFile;
//import ucar.nc2.Dimension;
import ucar.nc2.Variable;
import ucar.ma2.Array;
import ucar.ma2.Index;
import ucar.ma2.ArrayInt;
import ucar.ma2.ArrayDouble;
import ucar.ma2.DataType;

public class GridToImage implements ImageObserver {
	
	public static void main(String args[]) throws Exception {

		int outputImageWidth  = 600;
		int legendPanelWidth = 170;
		int outputImageHeight;
		int rMin = 255; // default red color component for max level
		int gMin = 255; // default red color component for max level
		int bMin = 255; // default red color component for max level
		int rMax =   0; // default red color component for min level
		int gMax =   0; // default red color component for min level
		int bMax =   0; // default red color component for min level
		int numClasses = 8;
        double max_value = -1;
		double values[] = null;
		String title = null;
        ArrayList<Annotation> annotations = new ArrayList<Annotation>();
        ArrayList<String> legendTitles = new ArrayList<String>();
        int legendPositionX = -25;
        int legendPositionY = -25;
		boolean drawLegend = true;
		boolean distributeEvenly = false;
        String legendPrintfFormat = "%f";
		double scaleFactor = 1;
		boolean gridSizeAnnotation = false;
		boolean renderCountsBlock = false;

		int ai = 0;
		while (ai < args.length && args[ai].startsWith("-")) {
            if (args[ai].equals("-w")) {
                ++ai;
                outputImageWidth = Integer.parseInt(args[ai++]);
			} else if (args[ai].equals("-lw")) {
                ++ai;
                legendPanelWidth = Integer.parseInt(args[ai++]);
			} else if (args[ai].equals("-n")) {
                ++ai;	
				numClasses = Integer.parseInt(args[ai++]);
			} else if (args[ai].equals("-h")) {
                ++ai;	
				renderCountsBlock = true;
			} else if (args[ai].equals("-m")) {
                ++ai;
                max_value = Double.parseDouble(args[ai++]);
			} else if (args[ai].equals("-s")) {
                ++ai;
				scaleFactor = Double.parseDouble(args[ai++]);
			} else if (args[ai].equals("-g")) {
                ++ai;
				gridSizeAnnotation = true;
			} else if (args[ai].equals("-N")) {
                ++ai;	
				numClasses = Integer.parseInt(args[ai++]);
				distributeEvenly = true;
			} else if (args[ai].equals("-C")) {
                ++ai;
				String valuesString = args[ai++];
				String stringValues[] = valuesString.split(",");
				int numValues = stringValues.length;
				values = new double[numValues];
				for (int i=0; i<numValues; ++i) {
					values[i] = Double.parseDouble(stringValues[i]);
				}
			} else if (args[ai].equals("-t")) {
                ++ai;	
				title = args[ai++];
			} else if (args[ai].equals("-nl")) {
                ++ai;	
				drawLegend = false;
			} else if (args[ai].equals("-a")) {
                ++ai;
                String text = args[ai++];
                int x = Integer.parseInt(args[ai++]);
                int y = Integer.parseInt(args[ai++]);
                annotations.add(new Annotation(text, x, y));
			} else if (args[ai].equals("-lt")) {
                ++ai;
                String text = args[ai++];
                legendTitles.add(text);
			} else if (args[ai].equals("-lp")) {
                ++ai;
                legendPositionX = Integer.parseInt(args[ai++]);
                legendPositionY = Integer.parseInt(args[ai++]);
			} else if (args[ai].equals("-lf")) {
                ++ai;
                legendPrintfFormat = args[ai++];
			} else if (args[ai].equals("-c")) {
                ++ai;
				rMin = Integer.parseInt(args[ai++]);
				gMin = Integer.parseInt(args[ai++]);
				bMin = Integer.parseInt(args[ai++]);
				rMax = Integer.parseInt(args[ai++]);
				gMax = Integer.parseInt(args[ai++]);
				bMax = Integer.parseInt(args[ai++]);
            } else {
                System.out.printf("unrecognized option: %s\n", args[ai]);
                System.exit(-1);
            }
        }

        String outputImageFile = args[ai++];
        String inputNetcdfFile = args[ai++];

		NetcdfFile ncfile = NetcdfFile.open(inputNetcdfFile);

        /*
        Variable latVar = ncfile.findVariable("lat");
        Variable lonVar = ncfile.findVariable("lon");
        */
        Variable latVar = ncfile.findVariable("y");
        Variable lonVar = ncfile.findVariable("x");
        ArrayDouble.D1 latData = (ArrayDouble.D1)(latVar.read());
        ArrayDouble.D1 lonData = (ArrayDouble.D1)(lonVar.read());
        double min_lat = latData.get(0);
        double max_lat = latData.get((int)(latData.getSize()-1));
        double min_lon = lonData.get(0);
        double max_lon = lonData.get((int)(lonData.getSize()-1));
		double cell_size = latData.get(1) - latData.get(0);

		outputImageHeight = (int)(Math.round(outputImageWidth * (max_lat - min_lat) / (max_lon - min_lon)));

		double grid_size_lat = latData.get(1) - min_lat;
		double grid_size_lon = lonData.get(1) - min_lon;

        GridToImage gti = new GridToImage();

		BufferedImage outputImage = new BufferedImage(outputImageWidth, outputImageHeight, BufferedImage.TYPE_INT_ARGB);

        Graphics2D g2d = (Graphics2D)outputImage.getGraphics();

        // 6. init map context and GTHelper
        //CoordinateReferenceSystem targetCRS = DefaultGeographicCRS.WGS84;
        CoordinateReferenceSystem targetCRS = CRS.parseWKT(SwdiGridder.ALBERS_WKT);
        MapContext context1 = new DefaultMapContext(targetCRS);
        MapContext context2 = new DefaultMapContext(targetCRS);

        GTHelper.addShapefileToContext(
                context1, 
                GTHelper.class.getResource("/shapefiles/countries-noUSA.shp"), 
                GTHelper.class.getResource("/styles/whiteBackground.sld"), 
                "Countries BG");
        
        GTHelper.addShapefileToContext(
                context1, 
                GTHelper.class.getResource("/shapefiles/states.shp"), 
                GTHelper.class.getResource("/styles/whiteBackground.sld"), 
                "States BG");
        
        GTHelper.addShapefileToContext(
                context1, 
                GTHelper.class.getResource("/shapefiles/lakes.shp"), 
                GTHelper.class.getResource("/styles/waterBackground.sld"), 
                "Lakes BG");

        ////////////////////////////////////////////////////////////////////////////////

        GTHelper.addShapefileToContext(
                context2, 
                GTHelper.class.getResource("/shapefiles/epsg3083/states_mask.shp"), 
                GTHelper.class.getResource("/styles/oceanBackground.sld"), 
                "States Mask");

        GTHelper.addShapefileToContext(
                context2, 
                GTHelper.class.getResource("/shapefiles/epsg3083/countries-noUSA.shp"), 
                GTHelper.class.getResource("/styles/whiteBackground.sld"), 
                "Countries BG");

        GTHelper.addShapefileToContext(
                context2, 
                GTHelper.class.getResource("/shapefiles/epsg3083/countries-noUSA.shp"), 
                GTHelper.class.getResource("/styles/countriesStyleFGMosaic.sld"), 
                "Countries FG");
        
        
        GTHelper.addShapefileToContext(
                context2, 
                GTHelper.class.getResource("/shapefiles/epsg3083/states.shp"), 
                GTHelper.class.getResource("/styles/stateStyleFGMosaic.sld"), 
                "States FG");

        StreamingRenderer renderer = new StreamingRenderer();
        ReferencedEnvelope refEnv  = new ReferencedEnvelope(min_lon, max_lon, min_lat, max_lat, targetCRS);

        renderer.setContext(context1);
        renderer.paint(
                (Graphics2D)outputImage.getGraphics(), 
                new Rectangle(outputImage.getWidth(), outputImage.getHeight()),
                GTHelper.adjustGeographicBounds(outputImage, refEnv));

        ColorRamp colorRamp =
			RenderGrid(outputImage, ncfile,
					   min_lat,
					   max_lat,
					   min_lon,
					   max_lon,
					   grid_size_lat,
					   grid_size_lon,
					   scaleFactor,
					   numClasses,
                       max_value,
					   values,
					   distributeEvenly,
					   rMin,
					   gMin,
					   bMin,
					   rMax,
					   gMax,
					   bMax);

        renderer.setContext(context2);
        renderer.paint(
                (Graphics2D)outputImage.getGraphics(), 
                new Rectangle(outputImage.getWidth(), outputImage.getHeight()),
                GTHelper.adjustGeographicBounds(outputImage, refEnv));


/*
        // Draw NOAA Logo
        URL localnoaaURL = MakeImage.class.getResource("/images/NOAA2small.png");
        BufferedImage noaaBI = ImageIO.read(localnoaaURL);
        g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 1.0f));
        g2d.drawImage(noaaBI, 1, outputImageHeight-noaaBI.getHeight()-1, null);
*/
        
		if (title != null) {
            Font font = new Font("Verdana", Font.BOLD, 16);
			g2d.setFont(font);
			FontRenderContext fc = g2d.getFontRenderContext();
			GlyphVector   glyphs = g2d.getFont().createGlyphVector(fc, title);
			Rectangle2D   bounds = glyphs.getVisualBounds();
        
			int textWidth = (int)bounds.getWidth();
			int textHeight = (int)bounds.getHeight();
			int xBuffer = 10;
			int yBuffer = 4;
			int bottomInsetHeight = 0;
        
			g2d.setColor(new Color(255, 255, 255, 128));
			g2d.fillRoundRect(outputImageWidth/2-(textWidth/2)-xBuffer,
							  //outputImageHeight-textHeight-yBuffer-bottomInsetHeight, 
							  bottomInsetHeight-yBuffer/2,
							  textWidth+(xBuffer*2), textHeight+(yBuffer*2), 5, 5);
        
			g2d.setColor(Color.BLACK);
			g2d.drawString(title, outputImageWidth/2-(textWidth/2), textHeight+bottomInsetHeight);
		}

        Font font = new Font("Verdana", Font.BOLD, 12);

        //Font font = g2d.getFont().deriveFont(g2d.getFont().getStyle(), 10);

		g2d.setFont( font );

		if (drawLegend) {

			int fullWidth = outputImageWidth+legendPanelWidth;
			int fullHeight = outputImageHeight;
			BufferedImage fullImage = new BufferedImage(fullWidth, fullHeight,
														BufferedImage.TYPE_INT_ARGB);
			Graphics2D fullg2d = (Graphics2D)(fullImage.getGraphics());
            Color oceanBlue = new Color(190,232,254);
			fullg2d.setColor(oceanBlue);
			fullg2d.fillRect(0,0,fullWidth,fullHeight);
			fullg2d.drawImage(outputImage,0,0,Color.WHITE,null);

			fullg2d.setFont( font );

            TextBlock legendTitleBlock = new TextBlock(font, 15);
            for (String legendTitle : legendTitles) {
                legendTitleBlock.appendLine(legendTitle);
            }

            ColorRampLegend legendRamp = new ColorRampLegend(colorRamp, 15, 0);

            TextBlock legendTextBlock = new TextBlock(font, 15);
            String text;
			double rampValues[] = colorRamp.getValues();
            for (int i=0; i<colorRamp.getNumClasses(); ++i) {
                if (i==0) {
                    text = String.format(String.format(" 1 - %s", legendPrintfFormat), rampValues[i+1]);
				} else if (i==colorRamp.getNumClasses()-1) {
                    text = String.format(String.format("  >= %s", legendPrintfFormat), rampValues[colorRamp.getNumClasses()-1]+1);
                } else {
                    text = String.format(String.format("%s - %s", legendPrintfFormat, legendPrintfFormat), rampValues[i]+1, rampValues[i+1]);
                }
                legendTextBlock.prependLine(text);
            }

            TextBlock countsBlock = new TextBlock(font, 15);
            int counts[] = colorRamp.getCounts();
            for (int i=0; i<colorRamp.getNumClasses(); ++i) {
                countsBlock.prependLine(String.format("[ %1d ]", counts[i]));
            }

            legendTitleBlock.render(fullg2d,
                               outputImageWidth+legendPanelWidth/2, 0,
                               0, 1, // anchorX, anchorY (anchor entire text block at middle of top edge)
                               0,    // lineAnchorX (center text on each line)
                               Color.BLACK);

            int legendHPad = 10;
            int legendVPad = 10;

            int y = legendTitleBlock.getHeight() + legendVPad;
            int x = outputImageWidth+legendPanelWidth/2 - (int)(Math.round( (legendTextBlock.getWidth() + legendHPad + legendRamp.getWidth()) / 2.0));
            legendTextBlock.render(fullg2d,
                              x, y-3,
                              -1, 1,
                              1,
                              Color.BLACK);
            x += legendTextBlock.getWidth() + legendHPad;
            legendRamp.render(fullg2d, x, y, -1, 1);

			if (renderCountsBlock) {
				countsBlock.render(fullg2d,
								   x + legendRamp.getWidth(),
								   y + legendRamp.getHeight() + 15,
								   1, 1,
								   1,
								   Color.BLACK);
			}

            outputImage = fullImage;
		}




		g2d = (Graphics2D)(outputImage.getGraphics());
        g2d.setColor(Color.BLACK);
		g2d.setFont( font );
        for (Annotation annotation : annotations) {
            annotation.render(g2d);
        }

		if (gridSizeAnnotation) {
            TextBlock gridSizeAnnotationTextBlock = new TextBlock(new Font("Verdana", Font.PLAIN, 12), 15);
			gridSizeAnnotationTextBlock.appendLine("grid cell size:");
			int rounded_cell_size_km = (int)Math.round(cell_size/1000.0);
			gridSizeAnnotationTextBlock.appendLine(String.format("%1dkm X %1dkm", rounded_cell_size_km, rounded_cell_size_km));
			gridSizeAnnotationTextBlock.render(g2d,
											   outputImageWidth+legendPanelWidth/2, outputImageHeight-10,
											   0, -1,
											   0,
											   Color.BLACK);
		}


		ImageIO.write(outputImage, "png", new File(outputImageFile));

		System.out.printf("gridconvert: wrote %s\n", outputImageFile);
	}


public static ColorRamp RenderGrid(BufferedImage bi, NetcdfFile ncfile,
								   double min_lat,
								   double max_lat,
								   double min_lon,
								   double max_lon,
								   double grid_size_lat,
								   double grid_size_lon,
								   double scaleFactor,
								   int numClasses,
								   double max_value,
								   double values[],
								   boolean distributeEvenly,
								   int rMin,
								   int gMin,
								   int bMin,
								   int rMax,
								   int gMax,
								   int bMax)
	throws Exception {
	
	/*
    private const double MilesPerDegLat = 69.04;
	private const double EarthRadiusMiles = 3963.1676;
	private static double PiDiv180 = Math.PI / 180;
	double MilesPerDegLon = MilesPerDegLat * Math.Cos(Latitude * PiDiv180)
	*/

	int outputImageWidth  = bi.getWidth();
	int outputImageHeight = bi.getHeight();

      Variable valueVar = ncfile.findVariable("value");

      Array valueData;
      DataType valueDataType = valueVar.getDataType();
      if (valueDataType == DataType.DOUBLE) {
          valueData = (ArrayDouble.D2) (valueVar.read());
      } else {
          valueData = (ArrayInt.D2) (valueVar.read());
      }
      int shape[] = valueData.getShape();

      double valueDouble[][] = new double[shape[1]][shape[0]];

      int width  = shape[1];
      int height = shape[0];

      Graphics2D g2d = bi.createGraphics();
      ImageCoordinateSystem biCoordinateSystem =
          new ImageCoordinateSystem(outputImageWidth, outputImageHeight, min_lon, min_lat, max_lon, max_lat);

      //double min_data_value = valueData.get(0,0);
      //double max_data_value = valueData.get(0,0);
      double min_data_value, max_data_value;
      Index valueIndex = valueData.getIndex();
        if (valueDataType == DataType.DOUBLE) {
            min_data_value = scaleFactor * valueData.getDouble(valueIndex.set(0,0));
            max_data_value = scaleFactor * valueData.getDouble(valueIndex.set(0,0));
        } else {
            min_data_value = scaleFactor * valueData.getInt(valueIndex.set(0,0));
            max_data_value = scaleFactor * valueData.getInt(valueIndex.set(0,0));
        }

        /*
         * Populate the valueDouble[][] array with the grid data.  As we do this,
         * keep track of the minimum and maximum data values, and the sum of
         * all the values (tot_val).
         */
        double tot_val = 0;
        double val;
        for (int i=0; i<width; ++i) {
            for (int j=0; j<height; ++j) {
                //int val = valueData.get(height-j-1,i);
                if (valueDataType == DataType.DOUBLE) {
                    val = scaleFactor * valueData.getDouble(valueIndex.set(j,i));
                } else {
                    val = scaleFactor * valueData.getInt(valueIndex.set(j,i));
                }
                valueDouble[i][j] = val;
                if (val > max_data_value) { max_data_value = val; }
                if (val < min_data_value) { min_data_value = val; }
                tot_val += val;
                //bi.setRGB(i,j,rgbint(255,255,255));
            }
        }
        System.out.printf("mm = %f, %f; tot=%f\n", min_data_value, max_data_value, tot_val);

		if (distributeEvenly) {

			int histogramResolution = 1000;
			int counts[] = new int[histogramResolution];
            //int counts[] = new int[((int)max_data_value)+1];

			double grand_tot = 0;
			for (int i=0; i<width; ++i) {
				for (int j=0; j<height; ++j) {
					//int val = valueData.get(height-j-1,i);
					if (valueDataType == DataType.DOUBLE) {
						val = scaleFactor * valueData.getDouble(valueIndex.set(j,i));
					} else {
						val = scaleFactor * valueData.getInt(valueIndex.set(j,i));
					}
					if (val > 0) {
						++counts[ (int)(val * (histogramResolution-1) / max_data_value) ];
						++grand_tot;
					}
				}
			}
			values = new double[numClasses+1];
			values[0] = 0;
			values[numClasses] = max_data_value;
			double cum_tot = 0;
			int classno = 1;
			for (int i=0; i<counts.length; ++i) {
				cum_tot += counts[i];
				if (cum_tot / grand_tot >= ((double)classno)/((double)numClasses)) {
					values[classno] = (i+1) * max_data_value / (histogramResolution-1);
					++classno;
					if (classno >= numClasses) { break; }
				}
			}

/*
System.out.printf("max_data_value = %f\n", max_data_value);
for (int i=0; i<counts.length; ++i) {
	System.out.printf("  %5d: %7d\n", i, counts[i]);
}
*/	

		}


        if (max_value <= 0) { max_value = max_data_value; }

		//ColorRamp colorRamp = new ColorRamp(0, max_value, numClasses, rMin, gMin, bMin, rMax, gMax, bMax);
		ColorRamp colorRamp;
		if (values == null) {
			colorRamp = new ColorRamp(0, max_value, numClasses, rMin, gMin, bMin, rMax, gMax, bMax);
		} else {
			colorRamp = new ColorRamp(values, rMin, gMin, bMin, rMax, gMax, bMax);
		}
		colorRamp.startCounts();
        for (int i=0; i<width; ++i) {
            for (int j=0; j<height; ++j) {
                if (valueDouble[i][j] > 0) {
					double lon0 = min_lon  + i * grid_size_lon;
					double lat0 = min_lat + j * grid_size_lat;
					ImageCoordinateSystem.IRect rect =
						biCoordinateSystem.coordsToPixel(new ImageCoordinateSystem.DRect(lon0,
																						 lat0,
																						 lon0 + grid_size_lon,
																						 lat0 + grid_size_lat));
					Color col = colorRamp.valueToColor(valueDouble[i][j]);
					g2d.setColor(col);
                    fillMinRect(g2d, rect.x0, rect.y1, rect.x1-rect.x0, rect.y0-rect.y1, 2, 2);
                }
            }
        }
		colorRamp.stopCounts();

		return colorRamp;
	}

    public static void fillMinRect(Graphics2D g2d, int x0, int y0, int w, int h, int minw, int minh) {
        if (w < minw) { w = minw; }
        if (h < minh) { h = minh; }
        g2d.fillRect(x0, y0, w, h);
    }

    public boolean imageUpdate(Image img, int infoflags, int x, int y, int width, int height)  {
        return false;
    }

}


		/*
        Graphics2D g2d = bi.createGraphics();
        ImageCoordinateSystem biCoordinateSystem =
            new ImageCoordinateSystem(outputImageWidth, outputImageHeight, min_lon, min_lat, max_lon, max_lat);

        g2d.setColor(new Color(255,255,255));
        g2d.fillRect(0,0,outputImageWidth,outputImageHeight);

		if (backgroundImageFile != null) {
			BufferedImage bgimg = ImageIO.read(new File(backgroundImageFile));
			System.out.printf("bg image is %1d X %1d\n", bgimg.getWidth(), bgimg.getHeight());
			ImageCoordinateSystem bgimgCoordinateSystem =
				new ImageCoordinateSystem(bgimg.getWidth(), bgimg.getHeight(), -125.0, 20.0, -67.0, 54.0);

			ImageCoordinateSystem.DRect bgimgExtent            = bgimgCoordinateSystem.getDRect();
			ImageCoordinateSystem.IRect bgimgDestPixels        = biCoordinateSystem.coordsToPixel(bgimgExtent);
			ImageCoordinateSystem.IRect clippedBgimgDestPixels = biCoordinateSystem.clip(bgimgDestPixels);
			ImageCoordinateSystem.DRect clippedBgimgExtent     = biCoordinateSystem.pixelToCoords(clippedBgimgDestPixels);
			ImageCoordinateSystem.IRect clippedBgimgSrcPixels  = bgimgCoordinateSystem.coordsToPixel(clippedBgimgExtent);

			g2d.drawImage(bgimg,
						clippedBgimgDestPixels.x0,
						clippedBgimgDestPixels.y0,
						clippedBgimgDestPixels.x1,
						clippedBgimgDestPixels.y1,
						clippedBgimgSrcPixels.x0,
						clippedBgimgSrcPixels.y0,
						clippedBgimgSrcPixels.x1,
						clippedBgimgSrcPixels.y1,
						gti);
		}
		*/
