/*
 * Author:
 *    Mark Phillips
 *    University of North Carolina Asheville /  National Climatic Data Center
 *    Mark.Phillips@noaa.gov
 *    828-271-4487
 */

package gov.noaa.ncdc.swdi.gridder;

public class ImageCoordinateSystem {

    public static class DPoint {
        public double x;
        public double y;
        public DPoint(double x, double y) {
            this.x = x;
            this.y = y;
        }
    }
    public static class DRect {
        public double x0, y0, x1, y1;
        public DRect(double x0, double y0, double x1, double y1) {
            this.x0 = x0;
            this.y0 = y0;
            this.x1 = x1;
            this.y1 = y1;
        }
        public DRect(DRect dRect) {
            this.x0 = dRect.x0;
            this.y0 = dRect.y0;
            this.x1 = dRect.x1;
            this.y1 = dRect.y1;
        }
        public String toString() {
            return String.format("[ ( %12.5f, %12.5f ),  ( %12.5f, %12.5f ) ]", x0, y0, x1, y1);
        }
    }
    public static class IPoint {
        public int x;
        public int y;
        public IPoint(int x, int y) {
            this.x = x;
            this.y = y;
        }
        public IPoint(double x, double y) {
            this.x = (int)(Math.round(x));
            this.y = (int)(Math.round(y));
        }
    }
    public static class IRect {
        public int x0, y0, x1, y1;
        public IRect(int x0, int y0, int x1, int y1) {
            this.x0 = x0;
            this.y0 = y0;
            this.x1 = x1;
            this.y1 = y1;
        }
        public IRect(double x0, double y0, double x1, double y1) {
            this.x0 = (int)(Math.round(x0));
            this.y0 = (int)(Math.round(y0));
            this.x1 = (int)(Math.round(x1));
            this.y1 = (int)(Math.round(y1));
        }
        public IRect(IRect iRect) {
            this.x0 = iRect.x0;
            this.y0 = iRect.y0;
            this.x1 = iRect.x1;
            this.y1 = iRect.y1;
        }
        public String toString() {
            return String.format("[ ( %4d, %4d ),  ( %4d, %4d ) ]", x0, y0, x1, y1);
        }
    }

    private int imageWidth, imageHeight;
    private int width, height;
    private double x0, y0, x1, y1;
    private double dx, dy;

    int getWidth()  { return this.width; }
    int getHeight() { return this.height; }
    double getX0()  { return this.x0; }
    double getY0()  { return this.y0; }
    double getX1()  { return this.x1; }
    double getY1()  { return this.y1; }

    public ImageCoordinateSystem(int width, int height, double x0, double y0, double x1, double y1) {
        this.imageWidth  = width;
        this.imageHeight = height;
        this.width  = width-1;
        this.height = height-1;
        this.x0     = x0;
        this.y0     = y0;
        this.x1     = x1;
        this.y1     = y1;
        this.dx     = x1 - x0;
        this.dy     = y1 - y0;
    }

    public IRect getIRect() {
        return new IRect(0, 0, width, height);
    }

    public DRect getDRect() {
        return new DRect(x0, y0, x1, y1);
    }

    public DPoint pixelToCoords(IPoint ip) {
        return new DPoint( x0 + dx*ip.x/width, y1 - dy*ip.y/height );
    }

    public DPoint pixelToCoords(int x, int y) {
        return new DPoint( x0 + dx*x/width, y1 - dy*y/height );
    }

    public DRect pixelToCoords(IRect iRect) {
        return new DRect( x0 + dx*iRect.x0/width,
                          y1 - dy*iRect.y0/height,
                          x0 + dx*iRect.x1/width,
                          y1 - dy*iRect.y1/height );
    }

    public IPoint coordsToPixel(DPoint dp) {
        return new IPoint( width * (dp.x - x0) / dx,  
                           height * ( 1 - (dp.y - y0)/dy ) );
    }

    public IPoint coordsToPixel(double x, double y) {
        return new IPoint( width * (x - x0) / dx,  
                           height * ( 1 - (y - y0)/dy ) );
    }

    public IRect coordsToPixel(DRect dRect) {
        return new IRect( width * (dRect.x0 - x0) / dx,  
                          height * ( 1 - (dRect.y0 - y0)/dy ),
                          width * (dRect.x1 - x0) / dx,  
                          height * ( 1 - (dRect.y1 - y0)/dy ) );
    }

    public IRect clip(IRect iRect) {
        IRect clippedIRect = new IRect(iRect);
        if (clippedIRect.x0 < 0) { clippedIRect.x0 = 0; }
        if (clippedIRect.x0 >= this.width) { clippedIRect.x0 = this.width-1; }
        if (clippedIRect.x1 < 0) { clippedIRect.x1 = 0; }
        if (clippedIRect.x1 >= this.width) { clippedIRect.x1 = this.width-1; }
        if (clippedIRect.y0 < 0) { clippedIRect.y0 = 0; }
        if (clippedIRect.y0 >= this.height) { clippedIRect.y0 = this.height-1; }
        if (clippedIRect.y1 < 0) { clippedIRect.y1 = 0; }
        if (clippedIRect.y1 >= this.height) { clippedIRect.y1 = this.height-1; }
        return clippedIRect;
    }

    public DRect clip(DRect dRect) {
        DRect clippedDRect = new DRect(dRect);
        if (clippedDRect.x0 < this.x0) { clippedDRect.x0 = this.x0; }
        if (clippedDRect.x0 > this.x1) { clippedDRect.x0 = this.x1; }
        if (clippedDRect.x1 < this.x0) { clippedDRect.x1 = this.x0; }
        if (clippedDRect.x1 > this.x1) { clippedDRect.x1 = this.x1; }
        if (clippedDRect.y0 < this.y0) { clippedDRect.y0 = this.y0; }
        if (clippedDRect.y0 > this.y1) { clippedDRect.y0 = this.y1; }
        if (clippedDRect.y1 < this.y0) { clippedDRect.y1 = this.y0; }
        if (clippedDRect.y1 > this.y1) { clippedDRect.y1 = this.y1; }
        return clippedDRect;
    }

}
