/*
 * Author:
 *    Mark Phillips
 *    University of North Carolina Asheville /  National Climatic Data Center
 *    Mark.Phillips@noaa.gov
 *    828-271-4487
 */

package gov.noaa.ncdc.swdi.gridder;

public class LatLonGridder {

    public static enum Alignment { LL, UL, CC }

    private DomainGridder latGridder, lonGridder;
    private Alignment alignment;

    public static class LatLon {
		public double lon, lat;
		public LatLon(double lon, double lat) {
			this.lon = lon;
			this.lat = lat;
		}
    }

    public static class GridIndex {
		public int row, col;
		public GridIndex(int col, int row) {
			this.row = row;
			this.col = col;
		}
    }

    public LatLonGridder(double west, double east, double south, double north, double gridSize, Alignment alignment, double precision) {
		this.alignment = alignment;
		switch (alignment) {
		case LL:
			latGridder = new DomainGridder(south, north, gridSize, DomainGridder.Alignment.MIN, precision);
			lonGridder = new DomainGridder(west,  east,  gridSize, DomainGridder.Alignment.MIN, precision);
			break;
		case UL:
			latGridder = new DomainGridder(south, north, gridSize, DomainGridder.Alignment.MAX, precision);
			lonGridder = new DomainGridder(west,  east,  gridSize, DomainGridder.Alignment.MIN, precision);
			break;
		case CC:
			latGridder = new DomainGridder(south, north, gridSize, DomainGridder.Alignment.CENTER, precision);
			lonGridder = new DomainGridder(west,  east,  gridSize, DomainGridder.Alignment.CENTER, precision);
			break;
		}
    }

    public static class OutOfBoundsException extends Exception {
		public OutOfBoundsException(String msg) {
			super(msg);
		}
		public OutOfBoundsException() {
			super();
		}
    }

    public GridIndex latLonToGridIndex(LatLon latLon) throws OutOfBoundsException {
		try {
			return new GridIndex( lonGridder.domainToIndex(latLon.lon),
								  latGridder.domainToIndex(latLon.lat) );
		} catch (DomainGridder.OutOfBoundsException e) {
			throw new OutOfBoundsException(e.getMessage());
		}
    }

    public GridIndex latLonToGridIndex(double lon, double lat) throws OutOfBoundsException {
		try {
			return new GridIndex( lonGridder.domainToIndex(lon),
								  latGridder.domainToIndex(lat) );
		} catch (DomainGridder.OutOfBoundsException e) {
			throw new OutOfBoundsException(e.getMessage());
		}
    }

    public LatLon gridIndexToLatLon(GridIndex index) throws OutOfBoundsException {
		try {
			return new LatLon( lonGridder.indexToDomain(index.col),
							   latGridder.indexToDomain(index.row) );
		} catch (DomainGridder.OutOfBoundsException e) {
			throw new OutOfBoundsException(e.getMessage());
		}
    }

    public LatLon gridIndexToLatLon(int col, int row) throws OutOfBoundsException {
		try {
			return new LatLon( lonGridder.indexToDomain(col),
							   latGridder.indexToDomain(row) );
		} catch (DomainGridder.OutOfBoundsException e) {
			throw new OutOfBoundsException(e.getMessage());
		}
    }

    public GridIndex getGridSize() {
		return new GridIndex( lonGridder.getCellCount(),
							  latGridder.getCellCount() );
    }

    public double indexToLat(int row) throws OutOfBoundsException {
		try {
			return latGridder.indexToDomain(row);
		} catch (DomainGridder.OutOfBoundsException e) {
			throw new OutOfBoundsException(e.getMessage());
		}
    }

    public double indexToLon(int col) throws OutOfBoundsException {
		try {
			return lonGridder.indexToDomain(col);
		} catch (DomainGridder.OutOfBoundsException e) {
			throw new OutOfBoundsException(e.getMessage());
		}
    }

}
