/*
 * Author:
 *    Mark Phillips
 *    University of North Carolina Asheville /  National Climatic Data Center
 *    Mark.Phillips@noaa.gov
 *    828-271-4487
 */

/*
 * For documentation on using this program, see the manual page "swdigridder.man", which
 * is located in the "man" subdirectory several levels above the directory containing
 * this source code.
 */

package gov.noaa.ncdc.swdi.gridder;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.ArrayList;
import java.io.File;
import java.io.Reader;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileReader;
import java.io.FileInputStream;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Statement;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Properties;
import java.util.TimeZone;
import java.util.zip.GZIPInputStream;

import ucar.nc2.*;
import ucar.ma2.*;

import org.geotools.geometry.jts.JTS;
import org.geotools.referencing.CRS;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.Coordinate;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.MathTransform;
import org.geotools.referencing.crs.DefaultGeographicCRS;

public class SwdiGridder {

    public static final String ALBERS_WKT = "PROJCS[\"NAD83 / Texas Centric Albers Equal Area\",GEOGCS[\"NAD83\",DATUM[\"North_American_Datum_1983\",SPHEROID[\"GRS 1980\",6378137,298.257222101],TOWGS84[0,0,0]],PRIMEM[\"Greenwich\",0,AUTHORITY[\"EPSG\",\"8901\"]],UNIT[\"degree\",0.01745329251994328,AUTHORITY[\"EPSG\",\"9122\"]],AUTHORITY[\"EPSG\",\"4269\"]],UNIT[\"metre\",1,AUTHORITY[\"EPSG\",\"9001\"]],PROJECTION[\"Albers_Conic_Equal_Area\"],PARAMETER[\"standard_parallel_1\",27.5],PARAMETER[\"standard_parallel_2\",35],PARAMETER[\"latitude_of_center\",18],PARAMETER[\"longitude_of_center\",-100],PARAMETER[\"false_easting\",1500000],PARAMETER[\"false_northing\",6000000],AUTHORITY[\"EPSG\",\"3083\"],AXIS[\"X\",EAST],AXIS[\"Y\",NORTH]]";

    public static class GridValue {
		public long period;
		public int count;
		public GridValue(long period, int count) {
			this.period = period;
			this.count = count;
		}
		public GridValue(long period) {
			this.period = period;
			this.count = 0;
		}
    }

	public static BufferedReader openPossiblyCompressedFile(String filename) throws Exception {
		if (filename.endsWith(".gz")) {
			return new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(new File(filename)))));
		}
		return new BufferedReader(new FileReader(new File(filename)));
	}

    public static void main(String[] args) throws Exception {

		//
		// usage:
		// 
		//     java SwdiGridder OUTPUT_FILE INPUT_FILE [...]
		//
		//  -s M                         (grid cell size, in 100ths of a degree; default is 5)
		//  -e WEST EAST SOUTH NORTH     (geographic bounding box; default is -125 -65 24 50)
		//  -b N                         (time binning period length, in minutes; default is 15.  Use -b 0 to disable time binning.)
        //  -i WSR_ID_FIELD[=WSR_ID ]   ...see man page...
		//  -f FILTER					 FILTER is an attribute filter expression consisting of one or more sequences like
		//									F o C
		//								 separated by '&&', where
		//									F is a column header name
		//									o is one of the operators <, <=, ==, !=, >, or >=
		//									C is either a number, or a string enclosed in single quotes
		//								 For example:
		//									WSR_ID != '??' && MAX_SHEAR > 10
		//								 Default is no filter.

		//double grid_cell_size = 0.05;
		double grid_cell_size = 5000;
        double xmin = -565205;
        double ymin = 6845734;
        double xmax = 4152350;
        double ymax = 9618240;
		int time_period_minutes = 15;
		String filter_string = null;
		String wsr_id_field = null;
		String wsr_id = null;
        String title = "SWDI Feature Frequency Plot";
        String value_long_name = "Number of features per grid cell";
        double value_divisor = 0.0;

		if (args.length < 2) {
			System.err.printf("\n");
			System.err.printf("usage: java -jar swdigridder.jar [OPTIONS] OUTPUT_FILE INPUT_FILE [...]\n");
			System.err.printf("\n");
			System.err.printf("See the file man/swdigridder.man for complete documentation.\n");
			System.err.printf("\n");
			System.exit(-1);
		}

		int ai = 0;
		while (args[ai].startsWith("-")) {
			if (args[ai].equals("-s")) {
				++ai;
				grid_cell_size = Double.parseDouble(args[ai++]);
            } else if (args[ai].equals("-d")) {
				++ai;
                value_divisor = Double.parseDouble(args[ai++]);
            } else if (args[ai].equals("-t")) {
				++ai;
                title = args[ai++];
            } else if (args[ai].equals("-l")) {
				++ai;
                value_long_name = args[ai++];
			} else if (args[ai].equals("-e")) {
				++ai;
				xmin = Double.parseDouble(args[ai++]); // west
				xmax = Double.parseDouble(args[ai++]); // east
				ymin = Double.parseDouble(args[ai++]); // south
				ymax = Double.parseDouble(args[ai++]); // north
			} else if (args[ai].equals("-b")) {
				++ai;
				time_period_minutes = Integer.parseInt(args[ai++]);
			} else if (args[ai].equals("-i")) {
				++ai;
				wsr_id_field = args[ai++];
				int ei = wsr_id_field.indexOf("=");
				if (ei >= 0) {
					String id_field = wsr_id_field.substring(0,ei);
					String id = wsr_id_field.substring(ei+1);
					wsr_id_field = id_field;
					wsr_id = id;
				}
				System.out.printf("wsr_id: %s", wsr_id_field);
				if (wsr_id != null) {
					System.out.printf("=%s", wsr_id);
				}
				System.out.println();
			} else if (args[ai].equals("-f")) {
				++ai;
				filter_string = args[ai++];
				System.out.println("filter: "+filter_string);
			} else {
				System.err.printf("unknown option: %s\n", args[ai]);
				System.exit(-1);
			}
		}

		String OUTPUT_FILE = args[ai++];

		if (!OUTPUT_FILE.endsWith(".nc")) {
			System.out.printf("output file name '%s' does not end with '.nc'.\n", OUTPUT_FILE);
		}

		ArrayList<String> inputFiles = new ArrayList<String>();
		while (ai < args.length) {
			inputFiles.add(args[ai++]);
		}

		//
		// Read the header from the first input file to get the column names
		//
		String firstInputFile = inputFiles.get(0);
		BufferedReader br = openPossiblyCompressedFile(firstInputFile);
		System.out.printf("reading column headers from file %s\n", firstInputFile);
		String line = br.readLine();
		while (line!=null && !line.startsWith("#ZTIME")) {
			line = br.readLine();
		}
    	//#ZTIME,LON,LAT,WSR_ID,CELL_ID,CELL_TYPE,RANGE,AZIMUTH,AVGDV,LLDV,MXDV,MXDV_HEIGHT,DEPTH,BASE,TOP,MAX_SHEAR,MAX_SHEAR_HEIGHT
		br.close();
		if (line == null) {
			System.out.printf("can't find header line in file %s\n", firstInputFile);
			System.exit(1);
		} else {
			line = line.replaceAll("#","");
			System.out.println("using headers: "+line);
		}
		String headers[] = line.split(",");

		// Put header fields in map and strip the leading '#'
		HashMap<String,Integer> columnIndexMap = new HashMap<String,Integer>();
		for (int i=0; i<headers.length; ++i) {
			columnIndexMap.put(headers[i].replaceAll("#", ""), i);
			//System.out.printf("%30s: %1d\n", headers[i], i);
		}

		//Filter filter = new Filter(columnIndexMap, filter_string);
		Filter filter = null;
		if (filter_string != null) {
			filter = new Filter(columnIndexMap, filter_string);
		}
		//System.out.printf("filter: %s\n", filter.toString());

		WsrIdMapper wsrIdMapper = new WsrIdMapper(wsr_id_field, wsr_id, inputFiles, columnIndexMap);

 		//
		// Allocate the internal grid array
		//
        /*
		double grid_min_y =  ymin;
        double d_yres = (ymax - ymin) / grid_cell_size;
        double floor_d_yres = Math.floor(d_yres);
        int yres = (
                    (floor_d_yres < d_yres)
                    ?
                    (int)(floor_d_yres + 1)
                    :
                    (int)floor_d_yres
                    );
        double grid_max_y = grid_min_y + yres * grid_cell_size;

		double grid_min_x =  xmin;
        double d_xres = (xmax - xmin) / grid_cell_size;
        double floor_d_xres = Math.floor(d_xres);
        int xres = (
                    (floor_d_xres < d_xres)
                    ?
                    (int)(floor_d_xres + 1)
                    :
                    (int)floor_d_xres
                    );
        double grid_max_x = grid_min_x + xres * grid_cell_size;

        System.out.printf("original extent is (%14.4f, %14.4f) to (%14.4f, %14.4f)\n",
                          xmin, ymin, xmax, ymax);
        System.out.printf("   final extent is (%14.4f, %14.4f) to (%14.4f, %14.4f)\n",
                          grid_min_x, grid_min_y, grid_max_x, grid_max_y);

		XYGridder gridder = new XYGridder(grid_min_x, grid_max_x, grid_min_y, grid_max_y,
                                          grid_cell_size, XYGridder.Alignment.LL, 0.0001);
        */
		XYGridder gridder = new XYGridder(xmin, xmax, ymin, ymax,
                                          grid_cell_size, XYGridder.Alignment.LL, 0.0001);
		XYGridder.GridIndex gridSize = gridder.getGridSize();
		int grid_width  = gridSize.col; // should equal xres
		int grid_height = gridSize.row; // should equal yres
		System.out.printf("allocating %d x %d grid...", grid_width, grid_height);
		GridValue grid[][] = new GridValue[grid_width][grid_height];
		System.out.printf("done.\n");

        //
        // Create the coordinate transform to convert lat/lon (WGS 84) to our desired coordinate system
        //
        CoordinateReferenceSystem targetCRS = CRS.parseWKT(ALBERS_WKT);
        MathTransform transform = CRS.findMathTransform(DefaultGeographicCRS.WGS84, targetCRS);
        GeometryFactory gf = new GeometryFactory();
        Point lonLatPoint, transformedPoint;
	    

		//
		// Read through all the input files, populating our internal feature array
		//
		ArrayList<SwdiFeature> swdiFeatures = new ArrayList<SwdiFeature>();

		String stringValues[] = null;
		double x, y, lon, lat;
		Date timestamp;
		SimpleDateFormat parser = new SimpleDateFormat("yyyyMMddHHmmss");
		int wsr_id_field_index = -1;
		if (wsr_id_field != null) {
			wsr_id_field_index = columnIndexMap.get(wsr_id_field);
		}

		for (String inputFile : inputFiles) {
			br = null;
			try {
				br = openPossiblyCompressedFile(inputFile);
				System.out.printf("reading data from file %s\n", inputFile);
				int lineno = 0;
				String wsrId;
				Coordinate coord = new Coordinate();
				String lonString, latString, timestampString;
				int i0, i1;
				while ( (line=br.readLine()) != null ) {
					++lineno;
					if (line.startsWith("#")) { continue; } // skip lines that start with #
					stringValues = line.trim().split(",");
					timestampString = stringValues[0];
					lonString = stringValues[1];
					latString = stringValues[2];

					/*
					i0 = 0;
					i1 = line.indexOf(',', i0);
					timestampString = line.substring(i0,i1);
					i0 = i1+1;
					i1 = line.indexOf(',', i0);
					lonString = line.substring(i0,i1);
					i0 = i1+1;
					i1 = line.indexOf(',', i0);
					latString = line.substring(i0,i1);
					*/

					if (filter!=null && !filter.evaluate(stringValues)) { continue; }

					try {
						coord.x = Double.parseDouble(lonString);
						coord.y = Double.parseDouble(latString);
                        lonLatPoint = gf.createPoint(coord);
                        transformedPoint = (Point)(JTS.transform( lonLatPoint, transform));

//System.out.printf("(lon,lat)=(%f,%f)  ==> (%f,%f)\n", lon,lat,transformedPoint.getCoordinate().x,transformedPoint.getCoordinate().y);

					} catch (Exception e) {
						System.out.printf("%s[%1d]: can't parse coordinates; line ignored.\n", inputFile, lineno);
						continue;
					}
					try {
						timestamp = parser.parse(timestampString);
					} catch (Exception e) {
						System.out.printf("%s[%1d]: can't parse timestamp; line ignored.\n", inputFile, lineno);
						continue;
					}

					wsrId = null;
					if (wsr_id_field_index >= 0) {
						wsrId = stringValues[ wsr_id_field_index ];
					}

					swdiFeatures.add( new SwdiFeature(transformedPoint.getCoordinate().x, transformedPoint.getCoordinate().y,
                                                      timestamp.getTime(), wsrId) );

					if (lineno > 0 && lineno % 100000 == 0) {
						System.out.printf("%1d lines read\n", lineno);
					}
				}

			} catch (Exception e) {
				System.out.println(e);
				e.printStackTrace();
			} finally {
				try { if (br != null) { br.close(); } }
				catch (Exception e) {
					System.out.println(e);
					e.printStackTrace();
				}
			}

		}

System.out.printf("total number of features read (after filtering) from input: %1d\n", swdiFeatures.size());

		Date baseTimestamp = parser.parse("19000101000000");
		long baseTimestampMillis = baseTimestamp.getTime();

		if (wsr_id_field == null) {
			createGridFile(OUTPUT_FILE,
						   grid_width, grid_height, grid, gridder,
						   baseTimestampMillis, time_period_minutes,
						   (SwdiFeature[])(swdiFeatures.toArray(new SwdiFeature[0])),
						   null, title, value_long_name, value_divisor);
		} else {
			if (wsr_id != null) {
				createGridFile(OUTPUT_FILE,
							   grid_width, grid_height, grid, gridder,
							   baseTimestampMillis, time_period_minutes,
							   (SwdiFeature[])(swdiFeatures.toArray(new SwdiFeature[0])),
							   wsr_id, title, value_long_name, value_divisor);
			} else {
				int k = OUTPUT_FILE.indexOf(".nc");
				String OUTPUT_FILE_FORMAT = OUTPUT_FILE.substring(0,k) + "-%s.nc";
				for (String the_wsr_id : wsrIdMapper.wsrIdArray) {
					createGridFile(String.format(OUTPUT_FILE_FORMAT, the_wsr_id),
								   grid_width, grid_height, grid, gridder,
								   baseTimestampMillis, time_period_minutes,
								   (SwdiFeature[])(swdiFeatures.toArray(new SwdiFeature[0])),
                                   wsr_id, title, value_long_name, value_divisor);
				}
			}

		}
			
		System.out.printf("done.\n");
			
	}
		
	public static void createGridFile(String OUTPUT_FILE,
									  int grid_width,
									  int grid_height,
									  GridValue grid[][],
									  XYGridder gridder,
									  long baseTimestampMillis,
									  int time_period_minutes,
									  SwdiFeature swdiFeatures[],
									  String wsr_id,
                                      String title,
                                      String value_long_name,
                                      double value_divisor) throws Exception {
		XYGridder.GridIndex idx = null;
		int max = -9999;
		long period;
		for (int i=0; i<grid_width; ++i) {
			for (int j=0; j<grid_height; ++j) {
				grid[i][j] = null;
			}
		}
        int featureCount = 0;
		for (SwdiFeature swdiFeature : swdiFeatures) {

			// TODO: determine whether to count this feature based on its WSR_ID here
			//if (this feature should not be counted because of its WSR_ID) { continue; }
			if (wsr_id!=null && !wsr_id.equals(swdiFeature.wsr_id)) { continue; }

			if (time_period_minutes > 0) {
				period = (long)(Math.floor( (double)(swdiFeature.timestampMillis - baseTimestampMillis) / (time_period_minutes*60*1000.0)));
			} else {
				period = -1;
			}

			try {
				idx = gridder.xyToGridIndex(swdiFeature.x, swdiFeature.y);
			} catch (XYGridder.OutOfBoundsException e) {
				//Silently continue (skipping this feature) on OutOfBounds exceptions;
				//it just means that this feature is outside the requested bounding box.
				continue;
			}

			try {
				if (grid[idx.col][idx.row] == null) {
					grid[idx.col][idx.row] = new GridValue(period, 1);
                    ++featureCount;
				} else {
					if ((period > grid[idx.col][idx.row].period) || (time_period_minutes <= 0)) {
						grid[idx.col][idx.row].period = period;
						++(grid[idx.col][idx.row].count);
                        ++featureCount;
					}
				}
				if (grid[idx.col][idx.row].count > max) { max = grid[idx.col][idx.row].count; }
			} catch (Exception e) {
				System.out.println(e);
				e.printStackTrace();
				System.out.printf("col=%1d, row=%1d, coords = (%f,%f)\n", idx.col, idx.row, swdiFeature.x, swdiFeature.y);
			}

		}

		//
		// Write the netcdf file
		//
		writeNetcdfFile(OUTPUT_FILE, grid_width, grid_height, grid, gridder,
                        title, value_long_name, value_divisor);

		System.out.printf("wrote %1d features to netcdf file %s\n", featureCount, OUTPUT_FILE);
	}

	public static void writeNetcdfFile(String OUTPUT_FILE, int grid_width, int grid_height, GridValue grid[][],
                                       XYGridder gridder,
                                       String title,
                                       String value_long_name,
                                       double value_divisor)
		throws Exception 
	{

		Index yIndex;
		Index xIndex;

		NetcdfFileWriteable ncfile;
		//ArrayInt valueArray;
        Array valueArray;
		Index valueIndex;

		Dimension yDim; 
		Dimension xDim;
		Dimension timeDim;
   
		ArrayDouble yArray;
		ArrayDouble xArray;

		ncfile = NetcdfFileWriteable.createNew(OUTPUT_FILE, true);

		yDim = ncfile.addDimension("y", grid_height);
		xDim = ncfile.addDimension("x", grid_width);

		ucar.nc2.Dimension[] dim2 = new ucar.nc2.Dimension[2];
		dim2[0] = yDim;
		dim2[1] = xDim;

        if (value_divisor > 0) {
            ncfile.addVariable("value", DataType.DOUBLE, dim2);
        } else {
            ncfile.addVariable("value", DataType.INT, dim2);
        }
		ncfile.addVariableAttribute("value", "long_name", value_long_name);
		ncfile.addVariableAttribute("value", "units",     "None");

		ncfile.addGlobalAttribute("title", title);
		//ncfile.addGlobalAttribute("grid cell side length in Albers Equal Area units"...note", "Each grid cell is %1d units on a side

		ncfile.addVariable("y", DataType.DOUBLE, new ucar.nc2.Dimension[] {yDim});
		ncfile.addVariableAttribute("y", "units", "degrees_north");

		ncfile.addVariable("x", DataType.DOUBLE, new ucar.nc2.Dimension[] {xDim});
		ncfile.addVariableAttribute("x", "units", "degrees_east");         

		ncfile.create();

        if (value_divisor > 0) {
            valueArray = new ArrayDouble.D2(yDim.getLength(), xDim.getLength());
        } else {
            valueArray = new ArrayInt.D2(yDim.getLength(), xDim.getLength());
        }
		valueIndex = valueArray.getIndex();

		yArray = new ArrayDouble.D1(yDim.getLength());
		yIndex = yArray.getIndex();
		xArray = new ArrayDouble.D1(xDim.getLength());
		xIndex = xArray.getIndex();
	    
		for (int i=0; i<grid_width; ++i) {
			double x = gridder.indexToX(i);
			xArray.setDouble(xIndex.set(i), x);
		}
	    
		for (int j=0; j<grid_height; ++j) {
			double y = gridder.indexToY(j);
			yArray.setDouble(yIndex.set(j), y);
		}
	    
		for (int i=0; i<grid_width; ++i) {
			for (int j=0; j<grid_height; ++j) {
				if (grid[i][j] == null) {
                    if (value_divisor > 0) {
                        valueArray.setDouble(valueIndex.set(j,i), 0.0);
                    } else {
                        valueArray.setInt(valueIndex.set(j,i), 0);
                    }
				} else {
                    if (value_divisor > 0) {
                        valueArray.setDouble(valueIndex.set(j,i), grid[i][j].count / value_divisor);
                    } else {
                        valueArray.setInt(valueIndex.set(j,i), grid[i][j].count);
                    }
				}
			}
		}

        {
            double vmax = 0;
            for (int i=0; i<grid_width; ++i) {
                for (int j=0; j<grid_height; ++j) {
                    if (grid[i][j]!=null && grid[i][j].count > vmax) { vmax = grid[i][j].count; }
                }
            }
            System.out.printf("grid max: %f; value_divisor = %f, quot = %f\n",
                              vmax, value_divisor, vmax / value_divisor);
        }


		ncfile.write("y", yArray);
		ncfile.write("x", xArray);
		ncfile.write("value", valueArray);
		ncfile.close();
	}


	public static class Filter {
		Expression expressions[];
		private enum Operator { Greater, GreaterEqual, Less, LessEqual, Equal, NotEqual, StringEqual, StringNotEqual };
		public String operatorToString(Operator op) {
			switch (op) {
			case Greater:			return ">";
			case GreaterEqual:		return ">=";
			case Less:				return "<";
			case LessEqual:			return "<=";
			case Equal:				return "==";
			case NotEqual:			return "!=";
			case StringEqual:		return "==";
			case StringNotEqual:	return "!=";
			}
			return "??";
		}
		public static class Expression {
			int index;
			Operator op;
			String stringRHS;
			double doubleRHS;
			public Expression(int index, Operator op, String stringRHS) {
				this.index = index;
				this.op = op;
				this.stringRHS = stringRHS;
			}
			public Expression(int index, Operator op, double doubleRHS) {
				this.index = index;
				this.op = op;
				this.doubleRHS = doubleRHS;
			}
			public boolean evaluate(String values[]) {
				switch (this.op) {
				case Greater:			return Double.parseDouble(values[index]) >  doubleRHS;
				case GreaterEqual:		return Double.parseDouble(values[index]) >= doubleRHS;
				case Less:				return Double.parseDouble(values[index]) <  doubleRHS;
				case LessEqual:			return Double.parseDouble(values[index]) <= doubleRHS;
				case Equal:				return Double.parseDouble(values[index]) == doubleRHS;
				case NotEqual:			return Double.parseDouble(values[index]) != doubleRHS;
				case StringEqual:		return values[index].equals(stringRHS);
				case StringNotEqual:	return !values[index].equals(stringRHS);
				}
				return true;
			}
			public String toString() {
				if (op == Operator.StringEqual || op == Operator.StringNotEqual) {
					return String.format("col[%1d] %s '%s'", index, op.toString(), stringRHS);
				}
				return String.format("col[%1d] %s %f", index, op.toString(), doubleRHS);
			}
		}
		public Filter(HashMap<String,Integer> columnIndexMap, String filter_string) throws Exception {
			if (filter_string == null || filter_string.equals("")) {
				this.expressions = new Expression[0];
				return;
			}
			String tokens[] = filter_string.trim().split("\\s+");
			ArrayList<Expression> expressions = new ArrayList<Expression>();
			int i=0;
			while (i<tokens.length) {
				String columnName = tokens[i++];
				if (!columnIndexMap.containsKey(columnName)) {
					throw new Exception(String.format("field '%s' not found in input file",columnName));
				}
				int index = columnIndexMap.get(columnName);
				String opString  = tokens[i++];
				String stringRHS = tokens[i++];
				double doubleRHS = 0;
				boolean rhsIsString = true;
				try {
					doubleRHS = Double.parseDouble(stringRHS);
					rhsIsString = false;
				} catch (Exception e) {}
				if (rhsIsString) {
					// remove quote delimiting string RHS, if present
					if (stringRHS.startsWith("'") && stringRHS.endsWith("'")) {
						stringRHS = stringRHS.substring(1,stringRHS.length()-1);
					}
				}
				Operator op = null;
				if (opString.equals(">"))  { op = Operator.Greater; }				
				else if (opString.equals(">=")) { op = Operator.GreaterEqual; }
				else if (opString.equals("<"))  { op = Operator.Less; }
				else if (opString.equals("<=")) { op = Operator.LessEqual; }
				else if (opString.equals("==")) { op = rhsIsString ? Operator.StringEqual : Operator.Equal; }
				else if (opString.equals("!=")) { op = rhsIsString ? Operator.StringNotEqual : Operator.NotEqual; }
				else {
					throw new Exception(String.format("unrecognized operator in filter expression: %s", opString));
				}
				if (rhsIsString) {
					expressions.add(new Expression(index, op, stringRHS));
				} else {
					expressions.add(new Expression(index, op, doubleRHS));
				}
				if (i < tokens.length) {
					if (!tokens[i++].equals("&&")) {
						throw new Exception("&& not found between filter expressions");
					}
				}
			}
			this.expressions = expressions.toArray(new Expression[] {});
		}
		public boolean evaluate(String values[]) {
			for (int i=0; i<expressions.length; ++i) {
				if (!expressions[i].evaluate(values)) { return false; }
			}
			return true;
		}
		public String toString() {
			if (expressions.length == 0) { return ""; }
			StringBuffer sb = new StringBuffer();
			sb.append(expressions[0].toString());
			for (int i=1; i<expressions.length; ++i) {
				sb.append(" && ");
				sb.append(expressions[i].toString());
			}
			return sb.toString();
		}
	}

	public static class WsrIdMapper {
		HashMap<String,Integer> wsrIdMap;
		public String wsrIdArray[];
		public String wsr_id_field, wsr_id;
		public enum Type { INDIVIDUAL, COMBINED, SINGLE };
		public Type getType() {
			if (wsr_id_field==null) { return Type.COMBINED; }
			if (wsr_id!=null) { return Type.SINGLE; }
			return Type.INDIVIDUAL;
		}
		public WsrIdMapper(String wsr_id_field, String wsr_id, List<String> inputFiles, HashMap<String,Integer> columnIndexMap) {
			this.wsr_id_field = wsr_id_field;
			this.wsr_id       = wsr_id;
			this.wsrIdMap     = null;
			this.wsrIdArray   = null;
			if (wsr_id_field != null && wsr_id == null) {
				wsrIdMap   = new HashMap<String,Integer>();
				BufferedReader br;
				String stringValues[];
				int index = 0;
				for (String inputFile : inputFiles) {
					br = null;
					try {
						br = openPossiblyCompressedFile(inputFile);
						String line;
						while ( (line=br.readLine()) != null ) {
							if (line.startsWith("#")) { continue; }
							stringValues = line.trim().split(",");
							String wsrId = stringValues[ columnIndexMap.get(wsr_id_field) ];
							if (! wsrIdMap.containsKey(wsrId) ) {
								wsrIdMap.put( wsrId, index++ );
							}
						}
						br.close();
					} catch (Exception e) {
						System.out.println(e);
						if (br != null) {
							try { br.close(); }
							catch (Exception e2) {}
						}
					}
				}
				if (wsrIdMap.size() != index) {
					System.out.printf("ERROR: wsrIdMap.size()=%1d but index=%1d\n", 
									  wsrIdMap.size(), index);
				}
				wsrIdArray = new String[wsrIdMap.size()];
				for (String wsrId : wsrIdMap.keySet()) {
					wsrIdArray[wsrIdMap.get(wsrId)] = wsrId;
				}

			}
		}

		public String indexToWsrId(int index) {
			if (wsrIdArray != null) { return wsrIdArray[index]; }
			return null;
		}

		public int wsrIdToIndex(String wsrId) {
			if (wsrIdMap != null) { return wsrIdMap.get(wsrId); }
			return 0;
		}


		public void test() {
			System.out.printf("WsrIdMapper.test:\n");
			if (wsrIdMap != null) {
				System.out.printf("  wsrIdMap.size() = %1d\n", wsrIdMap.size());
			}
			if (wsrIdArray != null) {
				System.out.printf("  wsrIdArray.length = %1d\n", wsrIdArray.length);
			}

			for (int i=0; i<wsrIdArray.length; ++i) {
				if (wsrIdMap.get(wsrIdArray[i]) != i) {
					System.out.printf("bad i=%1d\n", i);
				}
			}
			for (String wsrId : wsrIdMap.keySet()) {
				if (! wsrIdArray[ wsrIdMap.get(wsrId) ].equals( wsrId ) ) {
					System.out.printf("bad wsrId: %s\n", wsrId);
				}
			}
		}



	}

	public static class SwdiFeature {
		public double x;
		public double y;
		public long timestampMillis;
		public String wsr_id;
		public SwdiFeature (double x, double y, long timestampMillis, String wsr_id) {
			this.x       = x;
			this.y       = y;
			this.timestampMillis = timestampMillis;
			this.wsr_id          = wsr_id;
		}
	}



}
