package gov.noaa.ncdc.swdi.gridder;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.awt.geom.Rectangle2D;
import java.awt.FontMetrics;
import java.util.List;
import java.util.ArrayList;    
import java.awt.Font;

/**
 * A TextBlock is a list of strings which are to be rendered as a sequence
 * of lines of text, one under the other, in a Graphics2D object.  The class
 * provides control over the position and alignment of the entire text block
 * (the collection of all the lines of text, considered as a block), and the
 * horizontal alignment of the text on each line in the block.
 */
public class TextBlock {

    private Font font;
    private int lineHeight;
    private ArrayList<Annotation> annotations;

    private BufferedImage tmpImage;
    private Graphics2D tmpG2d;
    private FontMetrics fm;

    private int width;
    private int height;

    /**
     * Create a new TextBlock object using the given font.  If lineHeight >= 0
     * is specified, use it as the distance between the lines of text when rendered.
     * If lineHeight < 0, use the normal height of a line of text in the given font.
     */
    public TextBlock(Font font, int lineHeight) {
        this.font = font;
        this.lineHeight = lineHeight;
		this.tmpImage = new BufferedImage(400, 400, BufferedImage.TYPE_INT_ARGB);
        this.tmpG2d = (Graphics2D)(tmpImage.getGraphics());
        this.fm = tmpG2d.getFontMetrics();
        if (this.lineHeight < 0) {
            this.lineHeight = fm.getMaxAscent() + fm.getMaxDescent();
        }
        this.annotations = new ArrayList<Annotation>();
        this.width = 0;
        this.height = 0;
    }

    /**
     * Add a line of text to the bottom of the collection of lines to be rendered.
     */
    public void appendLine(String text) {
        addLine(text, 0);
    }

    /**
     * Add a line of text to the top of the collection of lines to be rendered.
     */
    public void prependLine(String text) {
        addLine(text, annotations.size());
    }

    /**
     * Return the current width of the entire text block, taking into account all
     * lines that have been added so far.
     */
    public int getWidth() { return this.width; }

    /**
     * Return the current height of the entire text block, taking into account all
     * lines that have been added so far.
     */
    public int getHeight() { return this.height; }



    /**
     * Render the text block in a Graphics2D.
     * <p>
     * The location of the text block is determined by a combination of two things:
     * <p>
     * <ul>
     *   <li>a "base" point (x,y) in the pixel coordinate system of the Graphics2D
     *   <li>an "anchor" point (anchorX,anchorY) in the text block, where (-1,-1)
     *       represents the lower left corner, and (1,1) represents the upper right.
     * </ul>
     * <p>
     * The text block is positioned so that its anchor point appears at the pixel
     * location (x,y) in the Graphics2D.  For example, if the anchor point is (-1,1),
     * the block will be position so that its upper left corner is at (x,y).  If
     * the anchor is (0,1), the middle of the text block's top edge will be at (x,y).
     * <p>
     * The alignment of the text on each line is governed by lineAlign, which should
     * be a number between -1 and 1.  The value -1 means left justify each line, 0 means
     * center each line, and 1 means right justify each line.
     * <p>
     * @param g2d   		the Graphics2D to render in
     * @param x				the x coordinate of the "base" point in g2d
     * @param y				the y coordinate of the "base" point in g2d
     * @param anchor X		the x coordinate of the "anchor" position in the text block
     * @param anchor Y		the y coordinate of the "anchor" position in the text block
     * @param lineAlign	    the horizontal alignment for the text on each line
     * @param color			the color to use for the text
     */
    public void render(Graphics2D g2d,
                       int x0, int y0,
                       double anchorX, double anchorY,
                       double lineAlign,
                       Color color) {
        Color oldColor = g2d.getColor();
		Font oldFont = g2d.getFont();
        int left   = (int)Math.round(x0 - (anchorX + 1)*this.width/2.0);
        int top    = (int)Math.round(y0 - (1 - anchorY)*this.height/2.0);
        int right  = left + width;
        int bottom = top + height;
        g2d.setColor(color);
		g2d.setFont(this.font);
		for (int i=0; i<annotations.size(); ++i) {
            try {
				Annotation ann = annotations.get(i);
				ann.y = bottom-(i)*this.lineHeight;
                int disp = (int)Math.round((lineAlign+1)*(this.width - ann.x)/2.0);
                ann.x = left + disp;
				ann.render(g2d);
            } catch (Exception e) {}
		}
        g2d.setColor(oldColor);
		g2d.setFont(oldFont);
    }

    private void addLine(String text, int index) {
        this.height += lineHeight;
        Rectangle2D   bounds = this.fm.getStringBounds(text, this.tmpG2d);
        int textWidth  = (int)(bounds.getWidth());
        int textHeight = (int)(bounds.getHeight());
        if (textWidth > this.width) {
            this.width = textWidth;
        }
        this.annotations.add(index, new Annotation(text, textWidth, textHeight));
    }

}
