/*
 * Author:
 *    Mark Phillips
 *    University of North Carolina Asheville /  National Climatic Data Center
 *    Mark.Phillips@noaa.gov
 *    828-271-4487
 */

package gov.noaa.ncdc.swdi.gridder;

public class XYGridder {

    public static enum Alignment { LL, UL, CC }

    private DomainGridder yGridder, xGridder;
    private Alignment alignment;

    public static class XY {
		public double x, y;
		public XY(double x, double y) {
			this.x = x;
			this.y = y;
		}
    }

    public static class GridIndex {
		public int row, col;
		public GridIndex(int col, int row) {
			this.row = row;
			this.col = col;
		}
    }

    public XYGridder(double west, double east, double south, double north, double gridSize, Alignment alignment, double precision) {
		this.alignment = alignment;
		switch (alignment) {
		case LL:
			yGridder = new DomainGridder(south, north, gridSize, DomainGridder.Alignment.MIN, precision);
			xGridder = new DomainGridder(west,  east,  gridSize, DomainGridder.Alignment.MIN, precision);
			break;
		case UL:
			yGridder = new DomainGridder(south, north, gridSize, DomainGridder.Alignment.MAX, precision);
			xGridder = new DomainGridder(west,  east,  gridSize, DomainGridder.Alignment.MIN, precision);
			break;
		case CC:
			yGridder = new DomainGridder(south, north, gridSize, DomainGridder.Alignment.CENTER, precision);
			xGridder = new DomainGridder(west,  east,  gridSize, DomainGridder.Alignment.CENTER, precision);
			break;
		}
    }

    public static class OutOfBoundsException extends Exception {
		public OutOfBoundsException(String msg) {
			super(msg);
		}
		public OutOfBoundsException() {
			super();
		}
    }

    public GridIndex xyToGridIndex(XY xy) throws OutOfBoundsException {
		try {
			return new GridIndex( xGridder.domainToIndex(xy.x),
								  yGridder.domainToIndex(xy.y) );
		} catch (DomainGridder.OutOfBoundsException e) {
			throw new OutOfBoundsException(e.getMessage());
		}
    }

    public GridIndex xyToGridIndex(double x, double y) throws OutOfBoundsException {
		try {
			return new GridIndex( xGridder.domainToIndex(x),
								  yGridder.domainToIndex(y) );
		} catch (DomainGridder.OutOfBoundsException e) {
			throw new OutOfBoundsException(e.getMessage());
		}
    }

    public XY gridIndexToXY(GridIndex index) throws OutOfBoundsException {
		try {
			return new XY( xGridder.indexToDomain(index.col),
							   yGridder.indexToDomain(index.row) );
		} catch (DomainGridder.OutOfBoundsException e) {
			throw new OutOfBoundsException(e.getMessage());
		}
    }

    public XY gridIndexToXY(int col, int row) throws OutOfBoundsException {
		try {
			return new XY( xGridder.indexToDomain(col),
							   yGridder.indexToDomain(row) );
		} catch (DomainGridder.OutOfBoundsException e) {
			throw new OutOfBoundsException(e.getMessage());
		}
    }

    public GridIndex getGridSize() {
		return new GridIndex( xGridder.getCellCount(),
							  yGridder.getCellCount() );
    }

    public double indexToY(int row) throws OutOfBoundsException {
		try {
			return yGridder.indexToDomain(row);
		} catch (DomainGridder.OutOfBoundsException e) {
			throw new OutOfBoundsException(e.getMessage());
		}
    }

    public double indexToX(int col) throws OutOfBoundsException {
		try {
			return xGridder.indexToDomain(col);
		} catch (DomainGridder.OutOfBoundsException e) {
			throw new OutOfBoundsException(e.getMessage());
		}
    }

}
